/*
**  NGPT - Next Generation POSIX Threading
**  Copyright (c) 2001 IBM Corporation <babt@us.ibm.com>
**  Portions Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
**
**  This file is part of NGPT, a non-preemptive thread scheduling
**  library which can be found at http://www.ibm.com/developer.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
**  USA.
**
**  pth_fork.c: Pth forking support
*/
                             /* ``Every day of my life
                                  I am forced to add another
                                  name to the list of people
                                  who piss me off!''
                                            -- Calvin          */
#include "pth_p.h"

struct pth_atfork_st {
    void (*prepare)(void *);
    void (*parent)(void *);
    void (*child)(void *);
    void *arg;
};

static struct pth_atfork_st pth_atfork_list[PTH_ATFORK_MAX];
static int pth_atfork_idx = 0;
intern int pth_fork_initialized = FALSE;

int pth_atfork_push(void (*prepare)(void *), void (*parent)(void *),
                    void (*child)(void *), void *arg)
{
    if (pth_atfork_idx > PTH_ATFORK_MAX-1)
        return FALSE;
    pth_atfork_list[pth_atfork_idx].prepare = prepare;
    pth_atfork_list[pth_atfork_idx].parent  = parent;
    pth_atfork_list[pth_atfork_idx].child   = child;
    pth_atfork_list[pth_atfork_idx].arg     = arg;
    pth_atfork_idx++;
    return TRUE;
}

int pth_atfork_pop(void)
{
    if (pth_atfork_idx <= 0)
        return FALSE;
    pth_atfork_idx--;
    return TRUE;
}

intern int pth_fork_init(void)
{
    int slot = 0;
    pth_attr_t t_attr;
    pth_descr_t descr = NULL;

    /* Initialize the native thread list... */
    for (slot = 0; slot < PTH_MAX_NATIVE_THREADS; slot++)
        memset(&pth_native_list[slot], 0x0, sizeof(struct pth_descr_st));
 
    /* initialize the scheduler */
    pth_scheduler_init();
 
    /* the current descriptor... */
    if ((descr = pth_get_native_descr()) == NULL) {
        fprintf(stderr,"pth_fork_init: unable to retrieve initial descriptor !?!?!?\n");
        abort();
    }
 
    /* spawn the scheduler thread */
    t_attr = pth_attr_new();
    if (t_attr == NULL) {
        fprintf(stderr,"pth_fork_init: unable to allocate initial attribute !?!?!?!\n");
        abort();
    }
    pth_attr_set(t_attr, PTH_ATTR_PRIO,         PTH_PRIO_MAX);
    pth_attr_set(t_attr, PTH_ATTR_NAME,         "**SCHEDULER**");
    pth_attr_set(t_attr, PTH_ATTR_JOINABLE,     FALSE);
    pth_attr_set(t_attr, PTH_ATTR_CANCEL_STATE, PTH_CANCEL_DISABLE);
    pth_attr_set(t_attr, PTH_ATTR_STACK_SIZE,   64*1024);
    pth_attr_set(t_attr, PTH_ATTR_STACK_ADDR,   NULL);
    descr->sched = pth_spawn(t_attr, pth_scheduler, NULL);
    if (descr->sched == NULL) {
        pth_attr_destroy(t_attr);
        close(pth_native_list[0].sigpipe[0]);
        close(pth_native_list[0].sigpipe[1]);
        abort();
    }
    descr->sched->lastrannative = descr->tid;
    pth_attr_destroy(t_attr);
    descr->current = descr->sched;

    return TRUE;
}

pid_t pth_fork(void)
{
    pid_t pid;
    int i;
    pth_descr_t descr;
    pth_t current;

    /* run preparation handlers in LIFO order */
    for (i = pth_atfork_idx-1; i >= 0; i--)
        if (pth_atfork_list[i].prepare != NULL)
            pth_atfork_list[i].prepare(pth_atfork_list[i].arg);

    /* above prepare routine might have changed our native, so check here... */
    if ((descr = pth_get_native_descr()) == NULL) {
	fprintf(stderr, "pth_fork: unable to find native task for pid %i.\n", gettid());
	return -1;
    }
    current = descr->current;

    /* fork the process */
    if ((pid = pth_sc(fork)()) == -1)
        return pid;

    /* handle parent and child contexts */
    if (pid != 0) {
        /* Parent: */

        /* run parent handlers in FIFO order */
        for (i = 0; i <= pth_atfork_idx-1; i++)
            if (pth_atfork_list[i].parent != NULL)
                pth_atfork_list[i].parent(pth_atfork_list[i].arg);
    }
    else {
        /* Child: */
	pth_pqueue_t *pth_CQ;

	/* update the descriptor values... */
	descr->pid	= getpid();
	descr->tid	= gettid();
	descr->current	= current;
	pth_fork_thread = current;
	pth_fork_initialized = TRUE;

	/* initialize global locks to avoid any deadlock */
	pth_lock_init(pth_key_lock);
	pth_lock_init(pth_usr1_lock);
	pth_lock_init(pth_native_lock);

        /* kick out all threads except for the current one and the scheduler */
        pth_fork_drop_threads();

        /* run child handlers in FIFO order */
        for (i = 0; i <= pth_atfork_idx-1; i++)
            if (pth_atfork_list[i].child != NULL)
                pth_atfork_list[i].child(pth_atfork_list[i].arg);

	/* drop the watchdog and native threads (other than currently running) */ 
	pth_fork_drop_natives();

	sigemptyset(&current->sigactionmask);
	if (current->sigpendcnt > 0) { 
		current->sigpendcnt = 0;
		sigemptyset(&current->sigpending);
	}

	/* initialize, create a new scheduler for this process */
	pth_fork_init();

	memset(&current->mctx, 0x0, sizeof(struct pth_mctx_st));
	descr = &pth_native_list[0];

	/* Add this thread to global thread list queue */
	_pth_acquire_lock(&pth_native_lock, descr->tid);
	pth_thread_insert(&current->thread_queue, &pth_threads_queue);
	pth_threads_count++;
	_pth_release_lock(&pth_native_lock, descr->tid);

	/* add current to pth_NQ list, so it can run from this scheduler */
	pth_CQ = &pth_NQ;
	current->state = PTH_STATE_NEW;
	_pth_acquire_lock(&(pth_CQ->q_lock), descr->tid);
	pth_pqueue_Qinsert(pth_CQ, current->prio, current);
	_pth_release_lock(&(pth_CQ->q_lock), descr->tid);

	/* Run this thread under new scheduler for this process */
	pth_mctx_switch(&current->mctx, &descr->sched->mctx);
	pth_main = current; /* set pth_main for this process */
	pth_fork_initialized = FALSE;
    }
    return pid;
}

intern void pth_fork_key_destroydata(pth_t t)
{
    /* Do not call the thread-specific destructor routine ... */

    pth_free_bytes(t->data_value, sizeof(void *)*PTH_KEY_MAX);
    t->data_value = NULL;
    return;
}

intern void pth_fork_tcb_free(pth_t t)
{
    if (t == NULL)
	return;

    /* Cleanup the stack. */
    if (t->true_stack != NULL && !t->stackloan) 
        munmap(t->true_stack, t->stacksize+8);

    /* free thread specific data */
    if (t->data_value != NULL)
        pth_fork_key_destroydata(t);

    /* free thread specific cleanups */
    if (t->cleanups != NULL)
        pth_cleanup_popall(t, FALSE);
    /*
     * If we adjusted the pointer previously to align on 16 byte boundary,
     * we need to put it back the proper way now.
     */
    if (t->ptrfixed)
        t = (pth_t)((char *)((char *)t - 8));
    /*ibm end*/

    munmap(t, sizeof(struct pth_st)+8);
    return;
}

intern void pth_fork_drop_natives(void)
{
    int     slot = 0;
    int     this_slot = 0;
    pid_t   tid = gettid();
 
    /* Signal the watchdog to terminate... */
    pth_watchdog_enabled = FALSE;

    /* Finish watchdog termination... */
    if (pth_watchdog_descr.is_used) {
        pth_watchdog_descr.is_used = FALSE;
        munmap(pth_watchdog_descr.true_stack, pth_watchdog_descr.stacksize);
    }
 
    do {
    	if (pth_native_list[slot].is_used && (tid != pth_native_list[slot].tid)) {
	 
        	/* Clean up the scheduler */
		if (pth_native_list[slot].current != pth_native_list[slot].sched)
			pth_fork_tcb_free(pth_native_list[slot].sched);
 
        	/* Now close the signal pipe... */
        	close(pth_native_list[slot].sigpipe[0]);
        	close(pth_native_list[slot].sigpipe[1]);
 
        	/*
         	 * Free clone stack...
        	 * an' finally, clean out the table entry...
         	 */
		if (slot != 0) { 
#if PTH_MCTX_MTH(sjlj)     &&\
    !PTH_MCTX_DSP(sjljlx)  &&\
    !PTH_MCTX_DSP(sjljisc) &&\
    !PTH_MCTX_DSP(sjljw32)
		   munmap(pth_native_list[slot].mctx_trampoline, sizeof(jmp_buf));
#endif
		   munmap(pth_native_list[slot].true_stack, pth_native_list[slot].stacksize);
		}
                pth_native_list[slot].is_used = FALSE;
	} else
		this_slot = slot;  /* save the current slot */
    } while (pth_native_list[++slot].is_used == TRUE);
 
    /* Now close the signal pipe for current native... */
    close(pth_native_list[this_slot].sigpipe[0]);
    close(pth_native_list[this_slot].sigpipe[1]);
    if (this_slot != 0)  
#if PTH_MCTX_MTH(sjlj)     &&\
    !PTH_MCTX_DSP(sjljlx)  &&\
    !PTH_MCTX_DSP(sjljisc) &&\
    !PTH_MCTX_DSP(sjljw32)
	   munmap(pth_native_list[this_slot].mctx_trampoline, sizeof(jmp_buf));
#endif
}
 
/* No locks are held (no contention here) */
intern void pth_fork_drop_threads(void)
{
   pth_t t;
   int slot = 0;

   /* clear the new queue */
    while ((t = pth_pqueue_delmax(&pth_NQ)) != NULL)
	pth_fork_tcb_free(t);
    pth_pqueue_init(&pth_NQ);

    /* clear the ready queue */
    while ((t = pth_pqueue_delmax(&pth_RQ)) != NULL)
	pth_fork_tcb_free(t);
    pth_pqueue_init(&pth_RQ);

    while ((t = pth_pqueue_delmax(&pth_WQ)) != NULL)
	pth_fork_tcb_free(t);
    pth_pqueue_init(&pth_WQ);

    /* clear the waiting queue */
    while (pth_native_list[slot].is_used) {
        pth_descr_t ds = &pth_native_list[slot++];
        while ((t = pth_pqueue_delmax(&(ds->wait_queue))) != NULL)
		pth_fork_tcb_free(t);
	pth_pqueue_init(&(ds->wait_queue));
    }

    /* clear the suspend queue */
    while ((t = pth_pqueue_delmax(&pth_SQ)) != NULL)
	pth_fork_tcb_free(t);
    pth_pqueue_init(&pth_SQ);

    /* clear the dead queue */
    while ((t = pth_pqueue_delmax(&pth_DQ)) != NULL)
	pth_fork_tcb_free(t);
    pth_pqueue_init(&pth_DQ);

    return;
}

